@@ -18,7 +18,7 @@ CURSOR_ITER_SIZE = 20000
 
 # pylint: disable=invalid-name,missing-function-docstring
 def calculate_destination_stream_name(stream, md_map):
-    return "{}-{}".format(md_map.get((), {}).get('schema-name'), stream['stream'])
+    return f"{md_map.get((), {}).get('schema-name')}-{stream['stream']}"
 
 
 # from the postgres docs:
@@ -29,59 +29,59 @@ def canonicalize_identifier(identifier):
 
 
 def fully_qualified_column_name(schema, table, column):
-    return '"{}"."{}"."{}"'.format(canonicalize_identifier(schema),
-                                   canonicalize_identifier(table),
-                                   canonicalize_identifier(column))
+    return f'"{canonicalize_identifier(schema)}"."{canonicalize_identifier(table)}"."{canonicalize_identifier(column)}"'
 
 
 def fully_qualified_table_name(schema, table):
-    return '"{}"."{}"'.format(canonicalize_identifier(schema), canonicalize_identifier(table))
+    return f'"{canonicalize_identifier(schema)}"."{canonicalize_identifier(table)}"'
 
 
 def open_connection(conn_config, logical_replication=False):
+    cfg = {
+        'application_name': 'pipelinewise',
+        'host': conn_config['host'],
+        'dbname': conn_config['dbname'],
+        'user': conn_config['user'],
+        'password': conn_config['password'],
+        'port': conn_config['port'],
+        'connect_timeout': 30
+    }
+
+    if conn_config.get('sslmode'):
+        cfg['sslmode'] = conn_config['sslmode']
+
     if logical_replication:
-        conn = psycopg2.connect(application_name='pipelinewise',
-                                host=conn_config['host'],
-                                dbname=conn_config['dbname'],
-                                user=conn_config['user'],
-                                password=conn_config['password'],
-                                port=conn_config['port'],
-                                connection_factory=psycopg2.extras.LogicalReplicationConnection,
-                                connect_timeout=30)
-    else:
-        conn = psycopg2.connect(application_name='pipelinewise',
-                                host=conn_config['host'],
-                                dbname=conn_config['dbname'],
-                                user=conn_config['user'],
-                                password=conn_config['password'],
-                                port=conn_config['port'],
-                                connect_timeout=30)
+        cfg['connection_factory'] = psycopg2.extras.LogicalReplicationConnection
+
+    conn = psycopg2.connect(**cfg)
 
     return conn
 
 def prepare_columns_for_select_sql(c, md_map):
-    column_name = ' "{}" '.format(canonicalize_identifier(c))
+    column_name = f' "{canonicalize_identifier(c)}" '
 
-    if ('properties', c) in md_map and md_map[('properties', c)]['sql-datatype'].startswith('timestamp'):
-        return f'CASE ' \
-               f'WHEN {column_name} < \'0001-01-01 00:00:00.000\' ' \
-               f'OR {column_name} > \'9999-12-31 23:59:59.999\' THEN \'9999-12-31 23:59:59.999\' ' \
-               f'ELSE {column_name} ' \
-               f'END AS {column_name}'
+    if ('properties', c) in md_map:
+        sql_datatype = md_map[('properties', c)]['sql-datatype']
+        if sql_datatype.startswith('timestamp') and not sql_datatype.endswith('[]'):
+            return f'CASE ' \
+                   f'WHEN {column_name} < \'0001-01-01 00:00:00.000\' ' \
+                   f'OR {column_name} > \'9999-12-31 23:59:59.999\' THEN \'9999-12-31 23:59:59.999\' ' \
+                   f'ELSE {column_name} ' \
+                   f'END AS {column_name}'
     return column_name
 
 def prepare_columns_sql(c):
-    column_name = """ "{}" """.format(canonicalize_identifier(c))
+    column_name = f""" "{canonicalize_identifier(c)}" """
     return column_name
 
 
 def filter_dbs_sql_clause(sql, filter_dbs):
-    in_clause = " AND datname in (" + ",".join(["'{}'".format(b.strip(' ')) for b in filter_dbs.split(',')]) + ")"
+    in_clause = " AND datname in (" + ",".join([f"'{b.strip(' ')}'" for b in filter_dbs.split(',')]) + ")"
     return sql + in_clause
 
 
 def filter_schemas_sql_clause(sql, filer_schemas):
-    in_clause = " AND n.nspname in (" + ",".join(["'{}'".format(b.strip(' ')) for b in filer_schemas.split(',')]) + ")"
+    in_clause = " AND n.nspname in (" + ",".join([f"'{b.strip(' ')}'" for b in filer_schemas.split(',')]) + ")"
     return sql + in_clause
 
 
@@ -150,11 +150,10 @@ def selected_value_to_singer_value_impl(elem, sql_datatype):
         if sql_datatype == 'hstore':
             cleaned_elem = elem
         else:
-            raise Exception("do not know how to marshall a dict if its not an hstore or json: {}".format(sql_datatype))
+            raise Exception(f"do not know how to marshall a dict if its not an hstore or json: {sql_datatype}")
     else:
         raise Exception(
-            "do not know how to marshall value of class( {} ) and sql_datatype ( {} )".format(elem.__class__,
-                                                                                              sql_datatype))
+            f"do not know how to marshall value of class( {elem.__class__} ) and sql_datatype ( {sql_datatype} )")
 
     return cleaned_elem
 
@@ -247,7 +246,7 @@ def numeric_min(precision, scale):
 
 
 def filter_tables_sql_clause(sql, tables: List[str]):
-    in_clause = " AND pg_class.relname in (" + ",".join(["'{}'".format(b.strip(' ')) for b in tables]) + ")"
+    in_clause = " AND pg_class.relname in (" + ",".join([f"'{b.strip(' ')}'" for b in tables]) + ")"
     return sql + in_clause
 
 def get_database_name(connection):
