@@ -7,10 +7,10 @@ import json
 import re
 import singer
 import warnings
-import singer.metadata as metadata
 
+from select import select
 from psycopg2 import sql
-from singer import utils, get_bookmark
+from singer import metadata, utils, get_bookmark
 from dateutil.parser import parse, UnknownTimezoneWarning, ParserError
 from functools import reduce
 
@@ -22,6 +22,7 @@ LOGGER = singer.get_logger('tap_postgres')
 
 UPDATE_BOOKMARK_PERIOD = 10000
 FALLBACK_DATETIME = '9999-12-31T23:59:59.999+00:00'
+FALLBACK_DATE = '9999-12-31T00:00:00+00:00'
 
 
 class ReplicationSlotNotFoundError(Exception):
@@ -59,7 +60,7 @@ def int_to_lsn(lsni):
         return None
 
     # Convert the integer to binary
-    lsnb = '{0:b}'.format(lsni)
+    lsnb = f'{lsni:b}'
 
     # file is the binary before the 32nd character, converted to hex
     if len(lsnb) > 32:
@@ -70,7 +71,7 @@ def int_to_lsn(lsni):
     # index is the binary from the 32nd character, converted to hex
     index = (format(int(lsnb[-32:], 2), 'x')).upper()
     # Formatting
-    lsn = "{}/{}".format(file, index)
+    lsn = f"{file}/{index}"
     return lsn
 
 
@@ -122,7 +123,7 @@ def get_stream_version(tap_stream_id, state):
     stream_version = singer.get_bookmark(state, tap_stream_id, 'version')
 
     if stream_version is None:
-        raise Exception("version not found for log miner {}".format(tap_stream_id))
+        raise Exception(f"version not found for log miner {tap_stream_id}")
 
     return stream_version
 
@@ -199,7 +200,7 @@ def create_array_elem(elem, sql_datatype, conn_info):
                 # custom datatypes like enums
                 cast_datatype = 'text[]'
 
-            sql_stmt = """SELECT $stitch_quote${}$stitch_quote$::{}""".format(elem, cast_datatype)
+            sql_stmt = f"""SELECT $stitch_quote${elem}$stitch_quote$::{cast_datatype}"""
             cur.execute(sql_stmt)
             res = cur.fetchone()[0]
             return res
@@ -211,6 +212,13 @@ def selected_value_to_singer_value_impl(elem, og_sql_datatype, conn_info):
 
     if elem is None:
         return elem
+
+    if sql_datatype == 'money':
+        return elem
+
+    if sql_datatype in ['json', 'jsonb']:
+        return json.loads(elem)
+
     if sql_datatype == 'timestamp without time zone':
         if isinstance(elem, datetime.datetime):
             # we don't want a datetime like datetime(9999, 12, 31, 23, 59, 59, 999999) to be returned
@@ -280,7 +288,15 @@ def selected_value_to_singer_value_impl(elem, og_sql_datatype, conn_info):
         if isinstance(elem, datetime.date):
             # logical replication gives us dates as strings UNLESS they from an array
             return elem.isoformat() + 'T00:00:00+00:00'
-        return parse(elem).isoformat() + "+00:00"
+        try:
+            return parse(elem).isoformat() + "+00:00"
+        except ValueError as e:
+            match = re.match(r'year (\d+) is out of range', str(e))
+            if match and int(match.group(1)) > 9999:
+                LOGGER.warning('datetimes cannot handle years past 9999, returning %s for %s',
+                               FALLBACK_DATE, elem)
+                return FALLBACK_DATE
+            raise
     if sql_datatype == 'time with time zone':
         # time with time zone values will be converted to UTC and time zone dropped
         # Replace hour=24 with hour=0
@@ -317,7 +333,7 @@ def selected_value_to_singer_value_impl(elem, og_sql_datatype, conn_info):
     if isinstance(elem, str):
         return elem
 
-    raise Exception("do not know how to marshall value of type {}".format(elem.__class__))
+    raise Exception(f"do not know how to marshall value of type {type(elem)}")
 
 
 def selected_array_to_singer_value(elem, sql_datatype, conn_info):
@@ -347,7 +363,7 @@ def row_to_singer_message(stream, row, version, columns, time_extracted, md_map,
 
         if not sql_datatype:
             LOGGER.info("No sql-datatype found for stream %s: %s", stream, columns[idx])
-            raise Exception("Unable to find sql-datatype for stream {}".format(stream))
+            raise Exception(f"Unable to find sql-datatype for stream {stream}")
 
         cleaned_elem = selected_value_to_singer_value(elem, sql_datatype, conn_info)
         row_to_persist += (cleaned_elem,)
@@ -380,7 +396,7 @@ def consume_message(streams, state, msg, time_extracted, conn_info):
     target_stream = streams_lookup[tap_stream_id]
 
     if payload['kind'] not in {'insert', 'update', 'delete'}:
-        raise UnsupportedPayloadKindError("unrecognized replication operation: {}".format(payload['kind']))
+        raise UnsupportedPayloadKindError(f"unrecognized replication operation: {payload['kind']}")
 
     # Get the additional fields in payload that are not in schema properties:
     # only inserts and updates have the list of columns that can be used to detect any different in columns
@@ -567,7 +583,7 @@ def sync_tables(conn_info, logical_streams, state, end_lsn, state_file):
     if version >= 120000:
         wal_sender_timeout = 10800000  # 10800000ms = 3 hours
         LOGGER.info('Set session wal_sender_timeout = %i milliseconds', wal_sender_timeout)
-        cur.execute("SET SESSION wal_sender_timeout = {}".format(wal_sender_timeout))
+        cur.execute(f"SET SESSION wal_sender_timeout = {wal_sender_timeout}")
 
     try:
         LOGGER.info('Request wal streaming from %s to %s (slot %s)',
@@ -585,107 +601,115 @@ def sync_tables(conn_info, logical_streams, state, end_lsn, state_file):
                               })
 
     except psycopg2.ProgrammingError as ex:
-        raise Exception("Unable to start replication with logical replication (slot {})".format(ex))
+        raise Exception(f"Unable to start replication with logical replication (slot {ex})") from ex
 
     lsn_received_timestamp = datetime.datetime.utcnow()
     poll_timestamp = datetime.datetime.utcnow()
 
-    while True:
-        # Disconnect when no data received for logical_poll_total_seconds
-        # needs to be long enough to wait for the largest single wal payload to avoid unplanned timeouts
-        poll_duration = (datetime.datetime.utcnow() - lsn_received_timestamp).total_seconds()
-        if poll_duration > logical_poll_total_seconds:
-            LOGGER.info('Breaking - %i seconds of polling with no data', poll_duration)
-            break
-
-        try:
-            msg = cur.read_message()
-        except Exception as e:
-            LOGGER.error(e)
-            raise
-
-        if msg:
-            if (break_at_end_lsn) and (msg.data_start > end_lsn):
-                LOGGER.info('Breaking - latest wal message %s is past end_lsn %s',
-                            int_to_lsn(msg.data_start),
-                            int_to_lsn(end_lsn))
+    try:
+        while True:
+            # Disconnect when no data received for logical_poll_total_seconds
+            # needs to be long enough to wait for the largest single wal payload to avoid unplanned timeouts
+            poll_duration = (datetime.datetime.utcnow() - lsn_received_timestamp).total_seconds()
+            if poll_duration > logical_poll_total_seconds:
+                LOGGER.info('Breaking - %i seconds of polling with no data', poll_duration)
                 break
 
             if datetime.datetime.utcnow() >= (start_run_timestamp + datetime.timedelta(seconds=max_run_seconds)):
                 LOGGER.info('Breaking - reached max_run_seconds of %i', max_run_seconds)
                 break
 
-            state = consume_message(logical_streams, state, msg, time_extracted, conn_info)
+            try:
+                msg = cur.read_message()
+            except Exception as e:
+                LOGGER.error(e)
+                raise
 
-            # When using wal2json with write-in-chunks, multiple messages can have the same lsn
-            # This is to ensure we only flush to lsn that has completed entirely
-            if lsn_currently_processing is None:
-                lsn_currently_processing = msg.data_start
-                LOGGER.info('First wal message received is %s', int_to_lsn(lsn_currently_processing))
+            if msg:
+                if (break_at_end_lsn) and (msg.data_start > end_lsn):
+                    LOGGER.info('Breaking - latest wal message %s is past end_lsn %s',
+                                int_to_lsn(msg.data_start),
+                                int_to_lsn(end_lsn))
+                    break
 
-                # Flush Postgres wal up to lsn comitted in previous run, or first lsn received in this run
-                lsn_to_flush = lsn_comitted
-                if lsn_currently_processing < lsn_to_flush:
-                    lsn_to_flush = lsn_currently_processing
-                LOGGER.info('Confirming write up to %s, flush to %s',
-                            int_to_lsn(lsn_to_flush),
-                            int_to_lsn(lsn_to_flush))
-                cur.send_feedback(write_lsn=lsn_to_flush, flush_lsn=lsn_to_flush, reply=True, force=True)
+                state = consume_message(logical_streams, state, msg, time_extracted, conn_info)
 
-            elif int(msg.data_start) > lsn_currently_processing:
-                lsn_last_processed = lsn_currently_processing
-                lsn_currently_processing = msg.data_start
-                lsn_received_timestamp = datetime.datetime.utcnow()
-                lsn_processed_count = lsn_processed_count + 1
-                if lsn_processed_count >= UPDATE_BOOKMARK_PERIOD:
-                    LOGGER.debug('Updating bookmarks for all streams to lsn = %s (%s)',
-                                 lsn_last_processed,
-                                 int_to_lsn(lsn_last_processed))
-                    for s in logical_streams:
-                        state = singer.write_bookmark(state, s['tap_stream_id'], 'lsn', lsn_last_processed)
-                    singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))
-                    lsn_processed_count = 0
+                # When using wal2json with write-in-chunks, multiple messages can have the same lsn
+                # This is to ensure we only flush to lsn that has completed entirely
+                if lsn_currently_processing is None:
+                    lsn_currently_processing = msg.data_start
+                    LOGGER.info('First wal message received is %s', int_to_lsn(lsn_currently_processing))
 
-        # Every poll_interval, update latest comitted lsn position from the state_file
-        if datetime.datetime.utcnow() >= (poll_timestamp + datetime.timedelta(seconds=poll_interval)):
-            if lsn_currently_processing is None:
-                LOGGER.info('Waiting for first wal message')
+                    # Flush Postgres wal up to lsn comitted in previous run, or first lsn received in this run
+                    lsn_to_flush = lsn_comitted
+                    if lsn_currently_processing < lsn_to_flush:
+                        lsn_to_flush = lsn_currently_processing
+                    LOGGER.info('Confirming write up to %s, flush to %s',
+                                int_to_lsn(lsn_to_flush),
+                                int_to_lsn(lsn_to_flush))
+                    cur.send_feedback(write_lsn=lsn_to_flush, flush_lsn=lsn_to_flush, reply=True, force=True)
+
+                elif int(msg.data_start) > lsn_currently_processing:
+                    lsn_last_processed = lsn_currently_processing
+                    lsn_currently_processing = msg.data_start
+                    lsn_received_timestamp = datetime.datetime.utcnow()
+                    lsn_processed_count = lsn_processed_count + 1
+                    if lsn_processed_count >= UPDATE_BOOKMARK_PERIOD:
+                        LOGGER.debug('Updating bookmarks for all streams to lsn = %s (%s)',
+                                    lsn_last_processed,
+                                    int_to_lsn(lsn_last_processed))
+                        for s in logical_streams:
+                            state = singer.write_bookmark(state, s['tap_stream_id'], 'lsn', lsn_last_processed)
+                        singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))
+                        lsn_processed_count = 0
             else:
-                LOGGER.info('Lastest wal message received was %s', int_to_lsn(lsn_last_processed))
                 try:
-                    state_comitted_file = open(state_file)
-                    state_comitted = json.load(state_comitted_file)
-                except Exception:
-                    LOGGER.debug('Unable to open and parse %s', state_file)
-                finally:
-                    lsn_comitted = min(
-                        [get_bookmark(state_comitted, s['tap_stream_id'], 'lsn') for s in logical_streams])
-                    if (lsn_currently_processing > lsn_comitted) and (lsn_comitted > lsn_to_flush):
-                        lsn_to_flush = lsn_comitted
-                        LOGGER.info('Confirming write up to %s, flush to %s',
-                                    int_to_lsn(lsn_to_flush),
-                                    int_to_lsn(lsn_to_flush))
-                        cur.send_feedback(write_lsn=lsn_to_flush, flush_lsn=lsn_to_flush, reply=True, force=True)
+                    # Wait for a second unless a message arrives
+                    select([cur], [], [], 1)
+                except InterruptedError:
+                    pass
 
-            poll_timestamp = datetime.datetime.utcnow()
+            # Every poll_interval, update latest comitted lsn position from the state_file
+            if datetime.datetime.utcnow() >= (poll_timestamp + datetime.timedelta(seconds=poll_interval)):
+                if lsn_currently_processing is None:
+                    LOGGER.info('Waiting for first wal message')
+                else:
+                    LOGGER.info('Lastest wal message received was %s', int_to_lsn(lsn_last_processed))
+                    try:
+                        with open(state_file, mode="r", encoding="utf-8") as fh:
+                            state_comitted = json.load(fh)
+                    except Exception:
+                        LOGGER.debug('Unable to open and parse %s', state_file)
+                    finally:
+                        lsn_comitted = min(
+                            [get_bookmark(state_comitted, s['tap_stream_id'], 'lsn') for s in logical_streams])
+                        if (lsn_currently_processing > lsn_comitted) and (lsn_comitted > lsn_to_flush):
+                            lsn_to_flush = lsn_comitted
+                            LOGGER.info('Confirming write up to %s, flush to %s',
+                                        int_to_lsn(lsn_to_flush),
+                                        int_to_lsn(lsn_to_flush))
+                            cur.send_feedback(write_lsn=lsn_to_flush, flush_lsn=lsn_to_flush, reply=True, force=True)
 
-    # Close replication connection and cursor
-    cur.close()
-    conn.close()
+                poll_timestamp = datetime.datetime.utcnow()
 
-    if lsn_last_processed:
-        if lsn_comitted > lsn_last_processed:
-            lsn_last_processed = lsn_comitted
-            LOGGER.info('Current lsn_last_processed %s is older than lsn_comitted %s',
-                        int_to_lsn(lsn_last_processed),
-                        int_to_lsn(lsn_comitted))
+        # Close replication connection and cursor
+        cur.close()
+        conn.close()
+    finally:
+        if lsn_last_processed:
+            if lsn_comitted > lsn_last_processed:
+                lsn_last_processed = lsn_comitted
+                LOGGER.info('Current lsn_last_processed %s is older than lsn_comitted %s',
+                            int_to_lsn(lsn_last_processed),
+                            int_to_lsn(lsn_comitted))
 
-        LOGGER.info('Updating bookmarks for all streams to lsn = %s (%s)',
-                    lsn_last_processed,
-                    int_to_lsn(lsn_last_processed))
+            LOGGER.info('Updating bookmarks for all streams to lsn = %s (%s)',
+                        lsn_last_processed,
+                        int_to_lsn(lsn_last_processed))
 
-        for s in logical_streams:
-            state = singer.write_bookmark(state, s['tap_stream_id'], 'lsn', lsn_last_processed)
+            for s in logical_streams:
+                state = singer.write_bookmark(state, s['tap_stream_id'], 'lsn', lsn_last_processed)
+
+        singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))
 
-    singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))
     return state
